#include "../../util/util.h"

using namespace std;

// texture identifiers
GLuint textureIDs[2];

static unsigned int	width, height;

// ping pong management vars
static bool	useZero = false;
static GLenum attachmentpoints[] = { GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT1_EXT };


// GLSL vars
static GLuint glslProgram;
static GLuint fragmentShader;
static GLint sourceTextureParam, alphaParam, hWidth, hHeight;

static float alpha;



PluginParam myParams[]={};

extern "C" int getParams(PluginParam** p){
	*p=myParams;
	return 0;//2 params
}

/**
 * swaps the role of the two y-textures (read-only and write-only)
 * Can be done in a smarter way :-)
 */
void swap(void) {
	useZero = !useZero;
}






/*
void createTextures()
{
    // create textures

    // y gets two textures, alternatingly read-only and write-only,
    // x is just read-only
    glGenTextures (2, textureIDs);
    // set up textures
    setupTexture (textureIDs[useZero]);
    transferToTexture(data,textureIDs[useZero]);
    setupTexture (textureIDs[!useZero]);
    transferToTexture(data,textureIDs[!useZero]);
    // set texenv mode from modulate (the default) to replace
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    // check if something went completely wrong
    checkGLErrors ("createFBOandTextures()");

}
*/















void	runShader()
{
	// TODO : accept the shader here, call loadShader()
	// also, accept dims

	// attach two textures to FBO
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, attachmentpoints[useZero], GL_TEXTURE_2D, textureIDs[useZero], 0);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, attachmentpoints[!useZero], GL_TEXTURE_2D, textureIDs[!useZero], 0);
	// check if that worked
	if (!checkFramebufferStatus()) {
		printf("glFramebufferTexture2DEXT():\t [FAIL]\n");
		exit (ERROR_FBOTEXTURE);
	}
	// enable GLSL program
	glUseProgram(glslProgram);
	
	printf("useProgram suxxeeded\n");

	// Calling glFinish() is only neccessary to get accurate timings,
	// and we need a high number of iterations to avoid timing noise.
	// glFinish();

	alpha = 2;
	glUniform1i(sourceTextureParam, 0);

	glUniform1f(alphaParam,alpha);
	glUniform1f(hWidth,width);
	glUniform1f(hHeight,height);

	
	for (int i=0; i<1; i++) {
		// set render destination
		glDrawBuffer (attachmentpoints[!useZero]);

		// enable texture y_old (read-only)
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D,textureIDs[useZero]);

		glUniform1i(sourceTextureParam, 0);

		// and render multitextured viewport-sized quad
		// depending on the texture target, switch between
		// normalised ([0,1]^2) and unnormalised ([0,w]x[0,h])
		// texture coordinates

		// make quad filled to hit every pixel/texel
		// (should be default but we never know)
		glPolygonMode(GL_FRONT,GL_FILL);

		// and render the quad
		// render with normalized texcoords
		glBegin(GL_QUADS);
		glTexCoord2f(0.0, 0.0);
		glVertex2f(0.0, 0.0);
		glTexCoord2f(1.0, 0.0);
		glVertex2f(width, 0.0);
		glTexCoord2f(1.0, 1.0);
		glVertex2f(width,height);
		glTexCoord2f(0.0, 1.0);
		glVertex2f(0.0, height);
		glEnd();
		// swap role of the two textures (read-only source becomes
		// write-only target and the other way round):
		swap();
		printf("swap done\n");
	}
	
	glBindTexture(GL_TEXTURE_2D,textureIDs[useZero]);
	
	// done, just do some checks if everything went smoothly.
	checkFramebufferStatus();
 	checkGLErrors("render()");

}

















extern "C" void run(GLuint	sourceTexID, int	_width, int	_height)
{
	// set textureIDs
	
	/*
	 * textureIDs is any length array, 0'th entry is always the source texture ID
	 * the plugin needs to make sure that the modified image is in the same tex
	 * ture before returning.
	 */
	
	textureIDs[useZero] = sourceTexID;
	width = _width;
	height = _height;
	
	
	
	
    // y gets two textures, alternatingly read-only and write-only, 
    glGenTextures (1, &(textureIDs[1]));
    // set up textures
    setupTexture (textureIDs[!useZero], width, height);
    // check if something went wrong
    checkGLErrors ("createFBOandTextures()");

	
	
	
    /**
     * initialize GLSL
     */

    // create program object
    glslProgram = glCreateProgram();
    // create shader object (fragment shader)
    fragmentShader = glCreateShader(GL_FRAGMENT_SHADER_ARB);
    // set source for shader
    const GLchar* source = \
    "#define KERNEL_SIZE 9\n"\
    "\n"\
    "// Gaussian kernel\n"\
    "// 1 2 1\n"\
    "// 2 4 2\n"\
    "// 1 2 1       \n"\
    "float kernel[KERNEL_SIZE];\n"\
    "\n"\
    "uniform sampler2D sourceTexture;\n"\
    "uniform float alpha;\n"\
    "uniform float width;\n"\
    "uniform float height;\n"\
    "\n"\
    "float step_w = 1.0/width;\n"\
    "float step_h = 1.0/height;\n"\
    "\n"\
    "vec2 offset[KERNEL_SIZE];\n"\
    "\n"\
    "void main(void)\n"\
    "{\n"\
    "       int i = 0;\n"\
    "       vec4 sum = vec4(0.0);\n"\
    "\n"\
    "       offset[0] = vec2(-step_w, -step_h);\n"\
    "       offset[1] = vec2(0.0, -step_h);\n"\
    "       offset[2] = vec2(step_w, -step_h);\n"\
    "\n"\
    "       offset[3] = vec2(-step_w, 0.0);\n"\
    "       offset[4] = vec2(0.0, 0.0);\n"\
    "       offset[5] = vec2(step_w, 0.0);\n"\
    "\n"\
    "       offset[6] = vec2(-step_w, step_h);\n"\
    "       offset[7] = vec2(0.0, step_h);\n"\
    "       offset[8] = vec2(step_w, step_h);\n"\
    "\n"\
    "\n"\
    "\n"\
    "	float	divisor = 1.0;\n"\
    "	float	colorOffset = 0.5;\n"\
//    "       kernel[0] = 1.0/16.0;   kernel[1] = 2.0/16.0;   kernel[2] = 1.0/16.0;\n"\
//    "       kernel[3] = 2.0/16.0;   kernel[4] = 4.0/16.0;   kernel[5] = 2.0/16.0;\n"\
//    "       kernel[6] = 1.0/16.0;   kernel[7] = 2.0/16.0;   kernel[8] = 1.0/16.0;\n"\

    "       kernel[0] = +4.0/divisor;   kernel[1] = -1.0/divisor;   kernel[2] = -1.0/divisor;\n"\
    "       kernel[3] = +1.0/divisor;   kernel[4] = +0.0/divisor;   kernel[5] = -1.0/divisor;\n"\
    "       kernel[6] = +1.0/divisor;   kernel[7] = +1.0/divisor;   kernel[8] = -4.0/divisor;\n"\

    "\n"\
    "\n"\
    "       for( i=0; i<KERNEL_SIZE; i++ )\n"\
    "       {\n"\
    "           vec4 tmp = texture2D(sourceTexture, gl_TexCoord[0].st + offset[i]);\n"\
    "           sum += tmp * kernel[i];\n"\
    "       }\n"\
    "\n"\
    "		vec4	color = clamp(sum + vec4(colorOffset), 0.0, 1.0);\n"\
    "		float	lum = 0.29*color.r +  0.59*color.g + 0.12*color.b;\n"\
    "       gl_FragColor = vec4(lum);\n"\
    "}\n";

    
    
    
    
    
    
    
    
    
    
    
    
    
//    " uniform sampler2D sourceTexture;\n"\
//    " void main(void)\n"\
//    " {\n"\
//    "    vec2 texCoord = gl_TexCoord[0].xy;\n"\
//    "    gl_FragColor = texture2D(sourceTexture, texCoord) * 0.5;\n"\
//    "    float dist;\n"\
//    "    for(dist = 1.0; dist <= 8.0; dist++) {\n"\
//    "    float offset = dist / 2500.0;\n"\
//    "\n"\
//    "\n"\
//    "    vec4 bl = texture2D(sourceTexture, texCoord + vec2(-offset, -offset));\n"\
//    "    vec4 l  = texture2D(sourceTexture, texCoord + vec2(-offset,     0.0));\n"\
//    "    vec4 tl = texture2D(sourceTexture, texCoord + vec2(-offset,  offset));\n"\
//    "    vec4 t  = texture2D(sourceTexture, texCoord + vec2(    0.0,  -offset));\n"\
//    "    vec4 ur = texture2D(sourceTexture, texCoord + vec2(    0.0,  offset));\n"\
//    "    vec4 r  = texture2D(sourceTexture, texCoord + vec2( offset,  -offset));\n"\
//    "    vec4 br = texture2D(sourceTexture, texCoord + vec2( offset,  0.0));\n"\
//    "    vec4 b  = texture2D(sourceTexture, texCoord + vec2( offset,  offset));\n"\
//    "	 vec4 finalColor = (bl + l + tl + t + ur + r + br + b);\n"\
//    "	 float intensity = (finalColor.s + finalColor.t + finalColor.p) / 3.0;\n"\
//    "    gl_FragColor = gl_FragColor + vec4(intensity / (64.0));\n"\
//    "   }\n"\
//    "  gl_FragColor = clamp(gl_FragColor, 0.0, 1.0);\n"\
//    " }\n";

    
    
    
    
    
//	    "uniform sampler2D sourceTexture;" \
//	    "uniform float alpha;" \
//	    "void main(void) { " \
//	    "  gl_FragColor = clamp(alpha * texture2D(sourceTexture, gl_TexCoord[0].st),0.0,1.0);"\
//	    "}";
    // clamp(alpha * texture2D(sourceTexture, gl_TexCoord[0].st),0.0,1.0);
    // clamp(alpha * texture2D(sourceTexture, gl_TexCoord[0].st),0.0,1.0);
    
    printf("sourcing ...\n");
    
    glShaderSource(fragmentShader, 1, &source, NULL);
    // compile shader
    glCompileShader(fragmentShader);
    printf("compile done\n");
    // check for errors
    printShaderInfoLog(fragmentShader);
    // attach shader to program
    glAttachShader (glslProgram, fragmentShader);
    printf("attach done\n");
    // link into full program, use fixed function vertex pipeline
    glLinkProgram(glslProgram);
    printf("link done\n");
    // check for errors
    printProgramInfoLog(glslProgram);
    checkGLErrors("render(2)");
    // Get location of the texture samplers for future use

    // Get location of the texture samplers for future use
    sourceTextureParam = glGetUniformLocation(glslProgram, "sourceTexture");
    alphaParam = glGetUniformLocation(glslProgram, "alpha");
    hWidth = glGetUniformLocation(glslProgram, "width");
    hHeight = glGetUniformLocation(glslProgram, "height");

    printf("setup done\n");
    runShader();
    
    return;
}



