#include "../../util/util.h"

using namespace std;

// texture identifiers
GLuint textureIDs[2];

static unsigned int	width, height;

// ping pong management vars
static bool	useZero = false;
static GLenum attachmentpoints[] = { GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT1_EXT };

// GLSL vars
static GLuint glslProgram1, glslProgram2;
static GLuint fragmentShader1,fragmentShader2;
static GLint sourceTextureParam1, radiusParam1, hWidth1, hHeight1;
static GLint sourceTextureParam2, radiusParam2, hWidth2, hHeight2;

static float radius;

PluginParam myParams[]={{"radius",1,ZERO_VEC4}};

/**
 * swaps the role of the two y-textures (read-only and write-only)
 * Can be done in a smarter way :-)
 */
void swap(void) {
	useZero = !useZero;
}

void	runShader()
{
	// TODO : accept the shader here, call loadShader()
	// also, accept dims

	// attach two textures to FBO
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, attachmentpoints[useZero], GL_TEXTURE_2D, textureIDs[useZero], 0);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, attachmentpoints[!useZero], GL_TEXTURE_2D, textureIDs[!useZero], 0);
	// check if that worked
	if (!checkFramebufferStatus()) {
		printf("glFramebufferTexture2DEXT():\t [FAIL]\n");
		exit (ERROR_FBOTEXTURE);
	}

//----

	
	for(int i = 0; i < 4; i++)
	{

	// enable GLSL program
	glUseProgram(glslProgram1);

	printf("useProgram1 suxxeeded\n");

	// Calling glFinish() is only neccessary to get accurate timings,
	// and we need a high number of iterations to avoid timing noise.
	// glFinish();


	glUniform1f(radiusParam1,radius / 4.0);
	glUniform1f(hWidth1,width);
	glUniform1f(hHeight1,height);

	// set render destination
	glDrawBuffer (attachmentpoints[!useZero]);

	// enable texture y_old (read-only)
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D,textureIDs[useZero]);

	glUniform1i(sourceTextureParam1, 0);

	// and render multitextured viewport-sized quad
	// depending on the texture target, switch between
	// normalised ([0,1]^2) and unnormalised ([0,w]x[0,h])
	// texture coordinates

	// make quad filled to hit every pixel/texel
	// (should be default but we never know)
	glPolygonMode(GL_FRONT,GL_FILL);

	// and render the quad
	// render with normalized texcoords
	glBegin(GL_QUADS);
	glTexCoord2f(0.0, 0.0);
	glVertex2f(0.0, 0.0);
	glTexCoord2f(1.0, 0.0);
	glVertex2f(width, 0.0);
	glTexCoord2f(1.0, 1.0);
	glVertex2f(width,height);
	glTexCoord2f(0.0, 1.0);
	glVertex2f(0.0, height);
	glEnd();
	// swap role of the two textures (read-only source becomes
	// write-only target and the other way round):
	swap();
	//printf("swap done\n");


// pass 2, horizontal

	
	// enable GLSL program
	glUseProgram(glslProgram2);

	printf("useProgram2 suxxeeded\n");

	// Calling glFinish() is only neccessary to get accurate timings,
	// and we need a high number of iterations to avoid timing noise.
	// glFinish();


	glUniform1f(radiusParam2,radius / 4.0 );
	glUniform1f(hWidth2,width);
	glUniform1f(hHeight2,height);

	// set render destination
	glDrawBuffer (attachmentpoints[!useZero]);

	// enable texture y_old (read-only)
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D,textureIDs[useZero]);

	glUniform1i(sourceTextureParam2, 0);

	// and render multitextured viewport-sized quad
	// depending on the texture target, switch between
	// normalised ([0,1]^2) and unnormalised ([0,w]x[0,h])
	// texture coordinates

	// make quad filled to hit every pixel/texel
	// (should be default but we never know)
	glPolygonMode(GL_FRONT,GL_FILL);

	// and render the quad
	// render with normalized texcoords
	glBegin(GL_QUADS);
	glTexCoord2f(0.0, 0.0);
	glVertex2f(0.0, 0.0);
	glTexCoord2f(1.0, 0.0);
	glVertex2f(width, 0.0);
	glTexCoord2f(1.0, 1.0);
	glVertex2f(width,height);
	glTexCoord2f(0.0, 1.0);
	glVertex2f(0.0, height);
	glEnd();
	// swap role of the two textures (read-only source becomes
	// write-only target and the other way round):
	swap();
	//printf("swap done\n");

	}

//----

	glBindTexture(GL_TEXTURE_2D,textureIDs[useZero]);

	// done, just do some checks if everything went smoothly.
	checkFramebufferStatus();
 	checkGLErrors("render()");

}

extern "C" int getParams(PluginParam** p){
	*p=myParams;
	return 1;//2 params
}

extern "C" void run(GLuint	sourceTexID, int	_width, int	_height)
{
	// set textureIDs

	/*
	 * textureIDs is any length array, 0'th entry is always the source texture ID
	 * the plugin needs to make sure that the modified image is in the same tex
	 * ture before returning.
	 */

	textureIDs[useZero] = sourceTexID;
	width = _width;
	height = _height;
	radius = myParams[0].param[0];

    // y gets two textures, alternatingly read-only and write-only,
    glGenTextures (1, &(textureIDs[1]));
    // set up textures
    setupTexture (textureIDs[!useZero], width, height);
    // check if something went wrong
    checkGLErrors ("createFBOandTextures()");

    /**
     * initialize GLSL
     */

    // create program objects
    glslProgram1 = glCreateProgram();
    glslProgram2 = glCreateProgram();
    // create shader object (fragment shader)
    fragmentShader1 = glCreateShader(GL_FRAGMENT_SHADER_ARB);
    fragmentShader2 = glCreateShader(GL_FRAGMENT_SHADER_ARB);

    // set source for shader
    const GLchar* source1 = \
    "uniform sampler2D sourceTexture;\n"\
    "uniform float radius;\n"\
    "uniform float width;\n"\
    "uniform float height;\n"\
    "\n"\
    "\n"\
    "void main(void)\n"\
    "{\n"\
    "       float i = 0.0;\n"\
    "       vec4 sum = vec4(0.0);\n"\
    "       vec2 offset = vec2(0.0);\n"\
    "       float weight = 1.0;\n"\
    "		float weightSum=0.0;\n"\
    "		\n"\
    "		float step_w = 1.0/width;\n"\
	"		float step_h = 1.0/height;\n"\
    "		float rad_by_1=1.0/radius;\n"
    "		float rad_by_12=rad_by_1*rad_by_1/2.0;\n"
	"		float rad_abs_i=0.0;"
    "\n"\
    "\n"\
    "       for( i=-radius; i<=radius; i+=1.0 )\n"\
    "       {\n"\
    "			offset=vec2(i*step_w,0.0);\n"\
	"			rad_abs_i=radius-abs(i);\n"\
    "			weight=(0.3989*rad_by_1)*exp(-rad_abs_i*rad_abs_i*rad_by_12);\n"\
    "			weightSum+=weight;\n"\
    "           vec4 tmp = texture2D(sourceTexture, gl_TexCoord[0].st + offset);\n"\
    "           sum += tmp*weight;\n"\
    "       }\n"\
    "		sum = sum/weightSum;\n"\
    "       gl_FragColor = clamp(sum,0.0,1.0);\n"\
    "}\n";


    const GLchar* source2 = \
    "uniform sampler2D sourceTexture;\n"\
    "uniform float radius;\n"\
    "uniform float width;\n"\
    "uniform float height;\n"\
    "\n"\
    "\n"\
    "void main(void)\n"\
    "{\n"\
    "       float i = 0.0;\n"\
    "       vec4 sum = vec4(0.0);\n"\
    "       vec2 offset = vec2(0.0);\n"\
    "       float weight = 1.0;\n"\
    "		float weightSum=0.0;\n"\
    "		\n"\
    "		float step_w = 1.0/width;\n"\
	"		float step_h = 1.0/height;\n"\
    "		float rad_by_1=1.0/radius;\n"
    "		float rad_by_12=rad_by_1*rad_by_1/2.0;\n"
	"		float rad_abs_i=0.0;"
    "\n"\
    "\n"\
    "       for( i=-radius; i<=radius; i+=1 )\n"\
    "       {\n"\
    "			offset=vec2(0.0,i*step_h);\n"\
	"			rad_abs_i=radius-abs(i);\n"\
    "			weight=(0.3989*rad_by_1)*exp(-rad_abs_i*rad_abs_i*rad_by_12);\n"\
    "			weightSum+=weight;\n"\
    "           vec4 tmp = texture2D(sourceTexture, gl_TexCoord[0].st + offset);\n"\
    "           sum += tmp*weight;\n"\
    "       }\n"\
    "		sum = sum/weightSum;\n"\
    "       gl_FragColor = clamp(sum,0.0,1.0);\n"\
    "}\n";


    printf("sourcing ...\n");

    glShaderSource(fragmentShader1, 1, &source1, NULL);
    glShaderSource(fragmentShader2, 1, &source2, NULL);
    // compile shader
    glCompileShader(fragmentShader1);
    glCompileShader(fragmentShader2);
    printf("compile done\n");
    // check for errors
    printShaderInfoLog(fragmentShader1);
    printShaderInfoLog(fragmentShader2);
    // attach shader to program
    glAttachShader (glslProgram1, fragmentShader1);
    glAttachShader (glslProgram2, fragmentShader2);
    printf("attach done\n");
    // link into full program, use fixed function vertex pipeline
    glLinkProgram(glslProgram1);
    glLinkProgram(glslProgram2);
    printf("link done\n");
    // check for errors
    printProgramInfoLog(glslProgram1);
    printProgramInfoLog(glslProgram2);
    checkGLErrors("render(2)");
    // Get location of the texture samplers for future use

    // Get location of the texture samplers for future use
    sourceTextureParam1 = glGetUniformLocation(glslProgram1, "sourceTexture");
    sourceTextureParam2 = glGetUniformLocation(glslProgram2, "sourceTexture");
    radiusParam1 = glGetUniformLocation(glslProgram1, "radius");
    radiusParam2 = glGetUniformLocation(glslProgram2, "radius");
    hWidth1 = glGetUniformLocation(glslProgram1, "width");
    hWidth2 = glGetUniformLocation(glslProgram2, "width");
    hHeight1 = glGetUniformLocation(glslProgram1, "height");
    hHeight2 = glGetUniformLocation(glslProgram2, "height");

    printf("setup done\n");
    runShader();

    return;
}


//"#define KERNEL_SIZE 9\n"\
//"\n"\
//"// Gaussian kernel\n"\
//"// 1 2 1\n"\
//"// 2 4 2\n"\
//"// 1 2 1\n"\
//"float kernel[KERNEL_SIZE];\n"\
//"\n"\
//"uniform sampler2D sourceTexture;\n"\
//"uniform float alpha;\n"\
//"uniform float width;\n"\
//"uniform float height;\n"\
//"\n"\
//"float step_w = 1.0/width;\n"\
//"float step_h = 1.0/height;\n"\
//"\n"\
//"vec2 offset[KERNEL_SIZE];\n"\
//"\n"\
//"void main(void)\n"\
//"{\n"\
//"       int i = 0;\n"\
//"       vec4 sum = vec4(0.0);\n"\
//"\n"\
//"       offset[0] = vec2(-step_w, -step_h);\n"\
//"       offset[1] = vec2(0.0, -step_h);\n"\
//"       offset[2] = vec2(step_w, -step_h);\n"\
//"\n"\
//"       offset[3] = vec2(-step_w, 0.0);\n"\
//"       offset[4] = vec2(0.0, 0.0);\n"\
//"       offset[5] = vec2(step_w, 0.0);\n"\
//"\n"\
//"       offset[6] = vec2(-step_w, step_h);\n"\
//"       offset[7] = vec2(0.0, step_h);\n"\
//"       offset[8] = vec2(step_w, step_h);\n"\
//"\n"\
//"\n"\
//"\n"\
//"	float	divisor = 16.0;\n"\
//"       kernel[0] = +1.0/divisor;   kernel[1] = +2.0/divisor;   kernel[2] = +1.0/divisor;\n"\
//"       kernel[3] = +2.0/divisor;   kernel[4] = +4.0/divisor;   kernel[5] = +2.0/divisor;\n"\
//"       kernel[6] = +1.0/divisor;   kernel[7] = +2.0/divisor;   kernel[8] = +1.0/divisor;\n"\
//
//"\n"\
//"\n"\
//"       for( i=0; i<KERNEL_SIZE; i++ )\n"\
//"       {\n"\
//"           vec4 tmp = texture2D(sourceTexture, gl_TexCoord[0].st + offset[i]);\n"\
//"           sum += tmp * kernel[i];\n"\
//"       }\n"\
//"\n"\
//"       gl_FragColor = clamp(sum,0.0,1.0);\n"\
//"}\n";
