#include "../../util/util.h"

using namespace std;

// texture identifiers
GLuint textureIDs[2];

static unsigned int	width, height;

// ping pong management vars
static bool	useZero = false;
static GLenum attachmentpoints[] = { GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT1_EXT };


// GLSL vars
static GLuint glslProgram;
static GLuint fragmentShader;
static GLint sourceTextureParam, whirlParam, radiusParam, centerParam, heightParam, widthParam;



PluginParam myParams[]={
	{"whirl amount",1,ZERO_VEC4},
	{"radius imcrement",1,ZERO_VEC4},
	{"whirl center",2,ZERO_VEC4},
};

/**
 * swaps the role of the two y-textures (read-only and write-only)
 * Can be done in a smarter way :-)
 */

void swap(void) {
	useZero = !useZero;
}

/*
void createTextures()
{
    // create textures

    // y gets two textures, alternatingly read-only and write-only,
    // x is just read-only
    glGenTextures (2, textureIDs);
    // set up textures
    setupTexture (textureIDs[useZero]);
    transferToTexture(data,textureIDs[useZero]);
    setupTexture (textureIDs[!useZero]);
    transferToTexture(data,textureIDs[!useZero]);
    // set texenv mode from modulate (the default) to replace
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    // check if something went completely wrong
    checkGLErrors ("createFBOandTextures()");

}
*/


void	runShader()
{
	// TODO : accept the shader here, call loadShader()
	// also, accept dims

	// attach two textures to FBO
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, attachmentpoints[useZero], GL_TEXTURE_2D, textureIDs[useZero], 0);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, attachmentpoints[!useZero], GL_TEXTURE_2D, textureIDs[!useZero], 0);
	// check if that worked
	if (!checkFramebufferStatus()) {
		printf("glFramebufferTexture2DEXT():\t [FAIL]\n");
		exit (ERROR_FBOTEXTURE);
	}
	// enable GLSL program
	glUseProgram(glslProgram);

	// Calling glFinish() is only neccessary to get accurate timings,
	// and we need a high number of iterations to avoid timing noise.
	// glFinish();

	// set render destination
	glDrawBuffer (attachmentpoints[!useZero]);
	// enable texture y_old (read-only)
	// MAY BE WRONG (PROLLY IS)
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D,textureIDs[useZero]);

	glUniform1i(sourceTextureParam, 0);
	glUniform1f(whirlParam,myParams[0].param[0]);
	glUniform1f(radiusParam,myParams[1].param[0]);
	glUniform2f(centerParam,myParams[2].param[0],myParams[2].param[1]);
	glUniform1i(heightParam,height);
	glUniform1i(widthParam,width);

	// and render multitextured viewport-sized quad
	// depending on the texture target, switch between
	// normalised ([0,1]^2) and unnormalised ([0,w]x[0,h])
	// texture coordinates

	// make quad filled to hit every pixel/texel
	// (should be default but we never know)
	glPolygonMode(GL_FRONT,GL_FILL);

	// and render the quad
	// render with normalized texcoords
	glBegin(GL_QUADS);
	glTexCoord2f(0.0, 0.0);
	glVertex2f(0.0, 0.0);
	glTexCoord2f(1.0, 0.0);
	glVertex2f(width, 0.0);
	glTexCoord2f(1.0, 1.0);
	glVertex2f(width,height);
	glTexCoord2f(0.0, 1.0);
	glVertex2f(0.0, height);
	glEnd();
	// swap role of the two textures (read-only source becomes
	// write-only target and the other way round):
	swap();


	glBindTexture(GL_TEXTURE_2D,textureIDs[useZero]);

	// done, just do some checks if everything went smoothly.
	checkFramebufferStatus();
 	checkGLErrors("render()");

}

extern "C" int getParams(PluginParam** p){
	*p=myParams;
	return 3;//3 params
}

extern "C" void run(GLuint	sourceTexID, int	_width, int	_height)
{
	// set textureIDs

	/*
	 * textureIDs is any length array, 0'th entry is always the source texture ID
	 * the plugin needs to make sure that the modified image is in the same tex
	 * ture before returning.
	 */

	textureIDs[useZero] = sourceTexID;
	width = _width;
	height = _height;

    // y gets two textures, alternatingly read-only and write-only,
    glGenTextures (1, &(textureIDs[1]));
    // set up textures
    setupTexture (textureIDs[!useZero], width, height);
    // check if something went wrong
    checkGLErrors ("createFBOandTextures()");


    /**
     * initialize GLSL
     */

    // create program object
    glslProgram = glCreateProgram();
    // create shader object (fragment shader)
    fragmentShader = glCreateShader(GL_FRAGMENT_SHADER_ARB);
    // set source for shader
    const GLchar* source = \
    "uniform	     sampler2D       sourceTexture ;\n"\
    "uniform         int   width;\n"\
    "uniform         int   height;\n"\
    "uniform         vec2  center;\n"\
    "uniform         float whirl;\n"\
    "uniform         float radius;\n"\
    "\n"\
    "void main()\n"\
    "{\n"\
    "        float   centerX = center.x / float(width);\n"\
    "        float   centerY = center.y / float(height);\n"\
    "        float   x = gl_TexCoord[0].x;\n"\
    "        float   y = gl_TexCoord[0].y;\n"\
    "        float   xsqPlusysq = (x-centerX)*(x-centerX)+(y-centerY)*(y-centerY);\n"\
    "        float   angle = whirl*exp(-(xsqPlusysq)/(radius*radius));\n"\
    "        float   u = cos(angle)*(x-centerX) + sin(angle)*(y-centerY);\n"\
    "        float   v = -sin(angle)*(x-centerX) + cos(angle)*(y-centerY);\n"\
    "        gl_FragColor = texture2D(sourceTexture, vec2(u+centerX,v+centerY));\n"\
    "}\n"\
    "\n";

    // clamp(alpha * texture2D(sourceTexture, gl_TexCoord[0].st),0.0,1.0);
    // clamp(alpha * texture2D(sourceTexture, gl_TexCoord[0].st),0.0,1.0);
    glShaderSource(fragmentShader, 1, &source, NULL);
    // compile shader
    glCompileShader(fragmentShader);
    // check for errors
    printShaderInfoLog(fragmentShader);
    // attach shader to program
    glAttachShader (glslProgram, fragmentShader);
    // link into full program, use fixed function vertex pipeline
    glLinkProgram(glslProgram);
    // check for errors
    printProgramInfoLog(glslProgram);
    checkGLErrors("render(2)");
    // Get location of the texture samplers for future use

    // Get location of the texture samplers for future use
    sourceTextureParam = glGetUniformLocation(glslProgram, "sourceTexture");
    whirlParam = glGetUniformLocation(glslProgram, "whirl");
    radiusParam = glGetUniformLocation(glslProgram, "radius");
    centerParam = glGetUniformLocation(glslProgram, "center");
    heightParam = glGetUniformLocation(glslProgram, "height");
    widthParam = glGetUniformLocation(glslProgram, "width");

    runShader();

    return;
}


