#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include <cstring>
#include <iostream>
#include <dlfcn.h>
#include <sys/types.h>
#include <dirent.h>
#include <vector>
#include "../util/util.h"

using namespace std;

char	*ppluginPath = NULL;

/* Structures for managing P-Plugins*/

struct PPlugin{
	char name[256];
};


//local list of p-plugins
vector<PPlugin> pPluginList;

/* Function prototypes */

static void query (void);
static void run (const gchar 	*name,
		gint        	nparams,
		const GimpParam *param,
		gint       	*nreturn_vals,
		GimpParam 	**return_vals);

extern int runGl(float * data, int _width, int _height);

/* Setting PLUG_IN_INFO */
GimpPlugInInfo PLUG_IN_INFO =
{
	NULL,    /* init_proc */
	NULL,    /* quit_proc */
	query,   /* query_proc */
	run     /* run_proc */
};

static void doStuff(const gchar* pluginName,GimpRunMode run_mode, guchar * buf_in, int width, int height, int bpp);

MAIN()

	int getPPlugins(char *directory) {
		DIR *pDIR;
		struct dirent *pDirEnt;

		/* Open the directory */

		pDIR = opendir(directory);

		if ( pDIR == NULL ) {
			printf("Error reading directory %s\n", directory);
			return -1;//return error
		}

		/* Get each directory entry from pDIR and print its name */

		pDirEnt = readdir( pDIR );//dir name

		while ( pDirEnt != NULL ) {
			//ignore anything without the words ".so"
			if (strstr(pDirEnt->d_name,".so")!=0){
				PPlugin *temp= new PPlugin;
				strcpy(temp->name,pDirEnt->d_name);
				char* found=strstr(temp->name,".so");//remove the letters .so
				*found='\0';
				printf( "Found Plugin: %s\n", temp);
				pPluginList.push_back(*temp);
			}
			pDirEnt = readdir( pDIR );
		}

		/* Release the open directory */

		closedir( pDIR );

		return 0;//return success
	}

static void query(void){
	//std::cout<<"GPU Shader Plugin Loaded";
	// printf("HOIME : %s",getenv("HOME"));
	static GimpParamDef params[] =
	{
		{ GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive" },
		{ GIMP_PDB_IMAGE, "image_id", "(unused)" },
		{ GIMP_PDB_DRAWABLE, "drawable", "Drawable to draw on" },
	};

	//static GimpParamDef ret_vals[] = {};

	char	*home = getenv("HOME");
	char	*pluginDir = "/.gpugimp";
	char	*path = new char[strlen(home) + strlen(pluginDir) + 1];
	strcpy(path, home);
	strcpy(path + strlen(home), pluginDir);
	ppluginPath = path;
	getPPlugins(path);

	for (int i=0;i<(int)pPluginList.size();i++){
		gimp_install_procedure(pPluginList.at(i).name, "GPU Shader",
				"Uses GPU to run a shader sequence image", "Duriel", "No rights reserved", "2008",
				pPluginList.at(i).name, "RGB", GIMP_PLUGIN, G_N_ELEMENTS(params),0,params,NULL);

		gimp_plugin_menu_register (pPluginList.at(i).name, "<Image>/Filters/GPU/");
	}
}

static void run (const gchar 	*name,
		gint        	nparams,
		const GimpParam *param,
		gint       	*nreturn_vals,
		GimpParam 	**return_vals){

	printf("\nui-run %u\n", rdtsc());

	//store status for returning to gimp (success/failure)
	GimpPDBStatusType  status = GIMP_PDB_SUCCESS;

	//store return values
	static GimpParam values[1];
	*nreturn_vals = 1;
	*return_vals  = values;

	//init the return status data to gimp
	values[0].type = GIMP_PDB_STATUS;
	values[0].data.d_status = status;

	GimpRunMode run_mode	= (GimpRunMode)param[0].data.d_int32;//get 0th parameter (run mode)
	//gint32 image_ID		= param[1].data.d_int32;//get 1st parameter (image id)

	//get 2nd parameter, which is the drawable we use.
	GimpDrawable *  drawable = gimp_drawable_get (param[2].data.d_drawable);


	//##FIXME: check for each pplugin we have and get its params etc...
	//if (strcmp (name, PROCEDURE_NAME) != 0 || nparams < 3){
	//	values[0].data.d_status = GIMP_PDB_CALLING_ERROR;
	//	return;
	//}

	if (true)//run_mode == GIMP_RUN_NONINTERACTIVE)
	{
		//code to get byte data and call doStuff

		//the pixel region used to read data
		GimpPixelRgn region_in;
		//the pixel region used to write data
		GimpPixelRgn region_out;

		//get the drawable's boundary
		gint x1, y1, x2, y2;
		gimp_drawable_mask_bounds (drawable->drawable_id, &x1, &y1, &x2, &y2);
		int width	=x2-x1;	//get width of drawable
		int height	=y2-y1;	//get height of drawable

		//get bytes per pixel of drawable
		int bpp		=gimp_drawable_bpp (drawable->drawable_id);

		//make a buffer to store original pixels, and one to write new pixel values
		guchar *buf_in 	= g_new (guchar, bpp*width*height);

		//init region to read from drawable (hence false,false)
		gimp_pixel_rgn_init	(&region_in, drawable, x1, y1, width, height, FALSE, FALSE);

		//init region to write to drawable (hence true,true)
		gimp_pixel_rgn_init	(&region_out, drawable, x1, y1, width, height, TRUE, TRUE);

		//read the region into a byte buffer.
		gimp_pixel_rgn_get_rect (&region_in, buf_in, x1, y1, width, height);

		gimp_progress_init ("Running Shaders...");

		printf("\nui-before-dostuff %u\n", rdtsc());

		//this is the function that takes buf_in as input, gives buf_out as output.
		//Note that both buffers have been pre allocated to width*height*bpp.
		doStuff(name,run_mode,buf_in,width,height,bpp);

		printf("\nui-after-dostuff %u\n", rdtsc());

		gimp_pixel_rgn_set_rect (&region_out, buf_in, x1, y1, width, height);

		g_free (buf_in);

		//end code to get byte data

		//clean up and update the drawable...
		gimp_drawable_flush(drawable);
		gimp_drawable_merge_shadow (drawable->drawable_id, TRUE);
		gimp_drawable_update (drawable->drawable_id, x1, y1,width,height);

		gimp_displays_flush ();
	}else{
		status = GIMP_PDB_CALLING_ERROR;
	}

	values[0].data.d_status = status;
	gimp_drawable_detach (drawable);

	printf("\nui-exit %u\n", rdtsc());
}

void parseParameters(float* param,int numParam,const char * text){
	char temp[256];
	strcpy(temp,text);
	char * pch;
	int count=0;

	pch = strtok (temp,", ");
	while (pch != NULL && count<numParam)
	{
		param[count]=atof(pch);
		pch = strtok (NULL, " ,.-");
		count++;
	}
}

//display the interface corresponding to the params
	bool makeInterface(const gchar * pluginName,PluginParam* pluginParameters,int numPluginParams){
		if (numPluginParams==0)
			return true;//dont show a dialog. no params needed.

		GtkWidget *dialog;
		vector<GtkWidget*> labels;
		vector<GtkWidget*> entries;
		gboolean   run;

		gimp_ui_init (pluginName, FALSE);

		dialog = gimp_dialog_new (pluginName, pluginName,
				NULL, (GtkDialogFlags)0,
				gimp_standard_help_func, "",
				GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				GTK_STOCK_OK,     GTK_RESPONSE_OK,
				NULL);

		GtkWidget * table = gtk_table_new(numPluginParams, 2, TRUE);
		gtk_table_set_row_spacings(GTK_TABLE(table), 2);
		gtk_table_set_col_spacings(GTK_TABLE(table), 2);

		int pos = 0;
		GtkWidget * tempWidget1, * tempWidget2;
		for(int i=0; i < numPluginParams; i++) {
			tempWidget1=gtk_label_new (pluginParameters[i].name);
			tempWidget2=gtk_entry_new ();
			labels.push_back(tempWidget1);
			entries.push_back(tempWidget2);
			gtk_table_attach_defaults(GTK_TABLE(table), tempWidget1, 0, 1, i, i+1 );
			gtk_table_attach_defaults(GTK_TABLE(table), tempWidget2, 1, 2, i, i+1 );
			pos++;
			cout<<"Added widgets"<<endl;
		}

		gtk_container_add(GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), table);
		gtk_widget_show_all(dialog);

		gtk_widget_show (dialog);

		run = (gimp_dialog_run (GIMP_DIALOG (dialog)) == GTK_RESPONSE_OK);

		if (run){
			for(int i=0; i < numPluginParams; i++) {
				parseParameters(pluginParameters[i].param,pluginParameters[i].type,gtk_entry_get_text ((GtkEntry*)entries.at(i)));
			}
		}

		gtk_widget_destroy (dialog);
		return run;
	}

//handles displaying the interface and other stuff...
bool handleGimpRunModes(const gchar * pluginName,PluginParam* pluginParameters,int numPluginParams,GimpRunMode run_mode){
	switch (run_mode){
		case	GIMP_RUN_INTERACTIVE:
			return makeInterface(pluginName,pluginParameters,numPluginParams);
		case	GIMP_RUN_NONINTERACTIVE:
			return makeInterface(pluginName,pluginParameters,numPluginParams);
		case	GIMP_RUN_WITH_LAST_VALS:
			//##FIXME: this should load and save shader parameters...
			return makeInterface(pluginName,pluginParameters,numPluginParams);
	}
	return false;
}

void doStuff(const gchar * pluginName,GimpRunMode run_mode, guchar * buf_in, int width, int height, int bpp){

	float	*data=new float[width*height*4];
	PluginParam *pluginParameters;
	int numPluginParams=0;

	printf("\nui-before-conv1 %u\n", rdtsc());
	for (int i=0;i<width*height;i++){
		data[i*4 +3]=buf_in[i*bpp+0]/255.0f;
		data[i*4 +2]=buf_in[i*bpp+1]/255.0f;
		data[i*4 +1]=buf_in[i*bpp+2]/255.0f;
		data[i*4 +0]=1.0f;//buf_in[i*bpp+3];
	}
	printf("\nui-after-conv1 %u\n", rdtsc());


	printf("\nui-before-glinit %u\n", rdtsc());
	glInit(data, width, height);
	printf("\nui-after-glinit %u\n", rdtsc());


	{
		// do the library stuff
		void *handle;
		void (*runPP)(GLuint	sourceTexID, int	_width, int	_height);
		int (*getParams)(PluginParam** p);
		char *error;
		char libPath[512];


		printf("constructing lib path");

		char	*home = getenv("HOME");
		char	*pluginDir = "/.gpugimp";
		char	*path = new char[strlen(home) + strlen(pluginDir) + 1];
		strcpy(path, home);
		strcat(path, pluginDir);
		ppluginPath = path;

		//construct the library name
		strcpy(libPath,ppluginPath);
		strcat(libPath,"/");
		strcat(libPath,pluginName);
		strcat(libPath,".so");

		printf(" lib path done\n");

		handle = dlopen(libPath, RTLD_NOW | RTLD_GLOBAL);
		if (!handle) {
			fprintf(stderr, "%s\n", dlerror());
			exit(EXIT_FAILURE);
		}

		dlerror();    /* Clear any existing error */

		*(void **) (&runPP) = dlsym(handle, "run");
		*(void **) (&getParams) = dlsym(handle, "getParams");

		if ((error = dlerror()) != NULL)  {
			fprintf(stderr, "%s\n", error);
			exit(EXIT_FAILURE);
		}
		printf("ytexid = %d.", yTexID);

		numPluginParams=getParams(&pluginParameters);

		if (handleGimpRunModes(pluginName,pluginParameters,numPluginParams,run_mode)){//user may press cancel

			printf("\nui-before-runpp %u\n", rdtsc());
			runPP(yTexID, width, height);
			printf("\nui-after-runpp %u\n", rdtsc());

		}
		dlclose(handle);

	}

	printf("\nui-before-transfer-from-tex %u\n", rdtsc());
	transferFromTexture(data);
	printf("\nui-after-transfer-from-tex %u\n", rdtsc());

	glDeInit();

	printf("\nui-before-conv2 %u\n", rdtsc());
	for (int i=0;i<width*height;i++){
		buf_in[i*bpp+0]=(guchar)(data[i*4 +3]*255);
		buf_in[i*bpp+1]=(guchar)(data[i*4 +2]*255);
		buf_in[i*bpp+2]=(guchar)(data[i*4 +1]*255);
	}
	printf("\nui-after-conv2 %u\n", rdtsc());

}

