#ifndef UTIL_H_
#define UTIL_H_

#include <cstdlib>
#include <iostream>
#include <GL/glew.h>
#include <GL/glut.h>

using namespace std;


// error codes
#define  ERROR_GLSL        -1
#define  ERROR_GLEW        -2
#define  ERROR_TEXTURE     -3
#define  ERROR_BINDFBO     -4
#define  ERROR_FBOTEXTURE  -5
#define  ERROR_PARAMS      -6

#define ZERO_VEC4 {0,0,0,0}

struct PluginParam{
	const char* name;
	int type;
	float param[4];
};

// texture identifiers
extern GLuint yTexID;

/**
 * Checks for OpenGL errors.
 * Extremely useful debugging function: When developing,
 * make sure to call this after almost every GL call.
 */
extern "C" void checkGLErrors (const char *label);

/**
 * Sets up a floating point texture with NEAREST filtering.
 */
extern "C" void setupTexture (const GLuint texID, int, int);

/**
 * Transfers data to texture.
 * Check web page for detailed explanation on the difference between ATI and NVIDIA.
 */
extern "C" void transferToTexture (float* data, int, int, GLuint texID);


/**
 * error checking for GLSL
 */
extern "C" void printProgramInfoLog(GLuint obj) ;
/**
 * prints out shader info, any errors during compilation etc.
 */
extern "C" void printShaderInfoLog(GLuint obj) ;

bool	glInit(float	*, int, int);


void glDeInit();



/**
 * Checks framebuffer status.
 * Copied directly out of the spec, modified to deliver a return value.
 */
bool checkFramebufferStatus();


/**
 * Transfers data from currently texture, and stores it in given array.
 */
void transferFromTexture(float* data);

/*
extern "C" void createTexture(GLuint	texID, int width, int height);
*/


/*on Intel(R) machines, reads the internal timestamp counter*/
static inline unsigned  rdtsc(void){
	static unsigned long long lastTime = 0;
	register unsigned long long tmp;
	__asm__ __volatile__ (
			"rdtsc"
			:"=A" (tmp)
			);
	unsigned long long diff = tmp - lastTime;
	lastTime = tmp;
	return (diff) / 2800000ULL;
}


#endif /*UTIL_H_*/
